
--[[

name:   Antix Level
vers:   1.0.1
desc:   Loads tiled maps and creates collision data from its objectlayers
auth:   by Cliff Earl
date:   March 2016
legal:  (c) 2016 Antix Software

Level will load and process all objectlayers in the supplied  Tiled map. For 
each object  it finds  with the  type  solid,  it will  create a   collision  
rectangle and add it to the  collider world. If Levels' createShapes flag is 
true, it will  also create a  new Gideros Shape  and add it to  itsself as a 
child.

When calling Level.new() you can supply an  optional table of options. Valid 
options are..

collider      - Hardon  Collider   world.  If  this  omitted,  no  collision 
                rectangles will be generated.
createShapes  - Bool flag. If true Level will create Gideros Shapes for each 
                corresponding collision rectangle created.
fillPen       - Default color for filling shapes
linePen       - Default color for outlining shapes



NOTE: There is placeholder code  where you can add functionality  to process 
      tilelayers and imagelayers.

--]]

Level = Core.class(Sprite)

function Level:init(options)
  self.rendered = false
  self.debug = true
  self.playerStart = {x = 0, y = 0}
  self.opacity = 1
  self.walls = {}
  self.mapName = ""
  self.map = nil

  if options then
    self.collider = options.collider or nil
    self.createShapes = options.createShapes or true
    self.fillPen = options.fillPen or 0x226644
    self.linePen = options.linePen or 0x339966
  else
    self.collider = self.nop
    self.createShapes = true
    self.fillPen = 0x226644
    self.linePen = 0x339966
  end  

end

function Level:log(text) -- DUMP TO CONSOLE
    if self.debug then print(text) end
end

function Level:nop()
  log("Collider not found!")
end

function Level:getPlayerStart() -- RETURN PLAYER START LOCATION
  return self.playerStart.x, self.playerStart.y
end

function Level:toFront()
  self:removeFromParent()
  stage:addChild(self)
end

function Level:loadMap(mapName, options) -- LOAD MAP, CREATE DISPLAY SHAPES, CREATE COLLISION DATA
  
  --
  -- DESTROY ANY PREVIOUSLY CREATED TILESETS AND TILEMAPS
  --
  
  if self.rendered then
    
    local childCount = self:getNumChildren() -- DESTROY ANY SHAPES
    if childCount > 0 then
      for i = childCount, 2, -1 do
        local shape = self:getChildAt(i)
        self:removeChildAt(i)
        shape:clear()
        shape = nil
      end
      self:removeChildAt(1)
    end

    if self.collider then
    end
    
    local walls = self.walls -- DESTROY COLLISION RECTANGLES
    for i = #walls, 1, -1 do
      local wall = walls[i]
      self.collider:remove(wall)
      table.remove(walls, i)
    end
  
    -- DESTROY OTHER STUFF HERE..
    
    self.rendered = false
  end
  
  self.mapName = "levels/"..mapName..".lua" -- LOAD THE MAP
  local func = loadfile(self.mapName)
  local map = func()
  func = nil
  self.map = map
  
  for i=1, #map.layers do
    local layer = map.layers[i]
    if layer.type == "tilelayer" then -- PROCESS TILE LAYER
      
      self:log("found tilelayer'"..layer.name.."'.. ignoring")
      
      --
      -- ADD CODE TO PROCESS TILE LAYERS IN HERE
      --
    
    elseif layer.type == "objectgroup" then -- PROCESS OBJECT LAYER HERE
      self.opacity = layer.opacity
      local objects = layer.objects
      for i = 1, #objects do
        local object = objects[i] -- PROCESS ALL OBJECTS IN LAYER
        if object.type == "playerstart" then
          self.playerStart = {x = object.x + (object.width * 0.5), y = object.y + (object.height * 0.5)}
        elseif object.type == "solid" then -- ONLY CREATE COLLISION SHAPES AND GRAPHICS FOR SOLID OBJECTS
          if object.shape == "rectangle" then
            self:newRect(object)
          elseif object.shape == "polygon" then
            self:newPolygon(object)
          elseif object.shape == "ellipse" then
            self:newCircle(object)
          else
            self:log("unknown shape '"..object.shape.."' in layer '"..layer.name.."'")
          end
          
        else
          self:log("unknown object type '"..object.type.."' in layer '"..layer.name.."'")
        
          --
          -- ADD CODE TO PROCESS TILE LAYERS IN HERE
          --
        
        end
      end
    elseif layer.type == "imagelayer" then -- PROCESS IMAGE LAYER HERE
      self:addChild(Bitmap.new(Texture.new(string.sub(layer.image, 4)))) -- LOAD BACKDROP
    else
      self:log("Level: found unknown layer type '"..layer.type.."'")
	  end
	end
  self.rendered = true
end

function Level:newCircle(object) -- CREATE NEW CIRCULAR SHAPE AND COLLISION DATA
  local d2r, sin, cos, insert = math.pi/180, math.sin, math.cos, table.insert
  
  local cx, cy, r = object.x + (object.width * 0.5), object.y + (object.height * 0.5), object.width * 0.5
  
  if self.createShapes then
    local circle = Shape.new() -- CREATE DISPLAY SHAPE
    self:setARGB(object, circle)
    circle:beginPath()
    local x1 = -r * sin(0)
    local y1 = r* cos(0)
    circle:moveTo(cx + -x1, cy + -y1)
    for c = 0, 360 - 6, 6 do
      x1 = -r * sin(c * d2r)
      y1 = r* cos(c * d2r)
      circle:lineTo(cx + -x1, cy + -y1)
    end
    circle:closePath()
    circle:endPath()
    self:addChild(circle)
  end
  
  local wall = self.collider:addCircle(cx, cy, r) -- CREATE COLLISION DATA
  wall.type = "tile"
  self.collider:addToGroup("tiles", wall)
  self.collider:setPassive(wall)
  insert(self.walls, wall)
end

function Level:newPolygon(object) -- CREATE NEW POLYGON SHAPE AND COLLISION DATA
  local insert = table.insert
  
  local x, y, polygon = object.x, object.y, object.polygon

  if self.createShapes then
    local poly = Shape.new() -- CREATE DISPLAY SHAPE
    self:setARGB(object, poly)
    poly:beginPath()
    poly:moveTo(x + polygon[1].x, y + polygon[1].y)
    for i = 2, #polygon do
      local vert = polygon[i]
      poly:lineTo(x + vert.x, y + vert.y)
    end
    poly:closePath()
    poly:endPath()
    self:addChild(poly)
  end
  
  local vertices = {} -- CREATE COLLISION DATA
  for i = 1, #polygon do
    local vert = polygon[i]
    insert(vertices, x + vert.x)
    insert(vertices, y + vert.y)
  end

  local wall = self.collider:addPolygon(unpack(vertices))
  wall.type = "tile"
  self.collider:addToGroup("tiles", wall)
  self.collider:setPassive(wall)
  insert(self.walls, wall)
end

function Level:newRect(object) -- CREATE NEW RECTANGLE SHAPE AND COLLISION DATA
  local insert = table.insert
  
  local x, y, w, h = object.x, object.y, object.width, object.height
  
  if self.createShapes then
    local rect = Shape.new() -- CREATE DISPLAY SHAPE
    self:setARGB(object, rect)
    rect:beginPath()
    rect:moveTo(x, y)
    rect:lineTo(x + w, y)
    rect:lineTo(x + w, y + h)
    rect:lineTo(x, y + h)
    rect:closePath()
    rect:endPath()
    self:addChild(rect)
  end
  
  local wall = self.collider:addPolygon(x, y, x + w, y, x + w, y + h, x, y + h) -- CREATE COLLISION DATA
  wall.type = "tile"
  self.collider:addToGroup("tiles", wall)
  self.collider:setPassive(wall)
  insert(self.walls, wall)
end

function Level:setARGB(object, shape) -- CREATE PENS FOR DRAWING SHAPE
  local function stringToRGB(text)
    local r = tonumber(string.sub(text, 1, 2), 16)
    local g = tonumber(string.sub(text, 3, 4), 16)
    local b = tonumber(string.sub(text, 5, 6), 16)
    return r * 65536 + g * 256 + b
  end
  
  local linePen = self.linePen
  local fillPen = self.fillPen
  local opacity = self.opacity
  if object.properties then
    local p = object.properties
    if p.opacity then opacity = tonumber(p.opacity) end
    if p.fillPen then fillPen = stringToRGB(p.fillPen) end
    if p.linePen then linePen = stringToRGB(p.linePen) end
  end
  shape:setFillStyle(Shape.SOLID, fillPen, opacity)
  shape:setLineStyle(2, linePen, opacity)
end
